/**
 * Copyright (c) 2011-2013 by Andrew Mustun. All rights reserved.
 *
 * This file is part of the QCAD project.
 *
 * QCAD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * QCAD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with QCAD.
 */

/*
 Strategy:
 
 - explode the selected entities
 - loop through all entites and try to convert them into polylines
 - sort polylines by bounding box into a tree (A contains B, etc.)
 - sort every branch by size
 - apply offsets by branch, flipping offset while walking down
 - possibly explode polylines back into individual lines
 - copy the remaining selected entities into the toolpath
 */

include("../Cam.js");
include("scripts/Block/Block.js");
include("scripts/ShapeAlgorithms.js");
//include("scripts/Pro/Draw/Polyline/PolylineFromSegments/PolylineFromSegments.js");

function Toolpath(guiAction) {
  Cam.call(this, guiAction);
}

Toolpath.prototype = new Cam();

Toolpath.getPreferencesCategory = function() {
  return [qsTr("Modify"), qsTr("Explode")];
};

Toolpath.prototype.beginEvent = function() {
  Cam.prototype.beginEvent.call(this);
  
  var di = this.getDocumentInterface();
  var document = this.getDocument();
  var storage = document.getStorage();
  var ids = document.querySelectedEntities();
  var i, k, e, n;
  var polyline, shapes, shape;
  
  var op = new RMixedOperation();
  var newShapes = new Array();
  var newPolylines = new Array();
  
  //di.deselectAll();
  
  // Take all selected entites and convert them into lines and arcs only
  // TODO: if we can concatenate Polylines, we don't need to convert them to lines and back
  for (i=0; i<ids.length; i++) {
    var id = ids[i];
    var entity = document.queryEntity(id);
    if (isNull(entity)) {
      debugger;
    }
    
    // delete entities that can not be cut
    if (isPointEntity(entity)) {
      continue;
    }
    
    // copy those entities that are already in the desired format
    else if (isLineEntity(entity)) {
      var cutLine = new RLine(entity.getStartPoint(), entity.getEndPoint());
      newShapes.push(cutLine);
    }
    else if (isArcEntity(entity)) {
    }
    /*
     var newEntities = new Array();
     
     // explode ellipse into polyline with arc segments:
     if (isEllipseEntity(entity)) {
     var ellipse = entity.getData().castToShape();
     var s = RSettings.getIntValue("Explode/EllipseSegments", 32);
     polyline = ShapeAlgorithms.approximateEllipse(ellipse, s);
     newShapes.push(polyline);
     }
     
     // explode polyline into line and arc segments:
     else if (isPolylineEntity(entity)) {
     polyline = entity.getData().castToShape();
     var polySegments = polyline.getExploded();
     if (polySegments.length>0) {
     for (k=0; k<polySegments.length; k++) {
     shape = polySegments[k].data();
     newShapes.push(shape);
     }
     }
     }
     
     // explode spline into polyline with line segments:
     else if (isSplineEntity(entity)) {
     var spline = entity.getData().castToShape();
     var s = RSettings.getIntValue("Explode/SplineSegments", 64);
     var lineSegments = spline.getExploded(s);
     if (lineSegments.length>0) {
     polyline = new RPolyline();
     polyline.appendVertex(lineSegments[0].getStartPoint());
     for (k=0; k<lineSegments.length; k++) {
     polyline.appendVertex(lineSegments[k].getEndPoint());
     }
     newShapes.push(polyline);
     }
     }
     
     // explode hatch into lines / solid fill into boundary:
     else if (isHatchEntity(entity)) {
     var painterPaths = entity.getPainterPaths(false);
     for (k=0; k<painterPaths.length; k++) {
     shapes = painterPaths[k].getShapes();
     for (n=0; n<shapes.length; n++) {
     shape = shapes[n].data();
     if (isSplineShape(shape)) {
     shape = ShapeAlgorithms.splineToLineOrArc(shape, 0.01);
     }
     newShapes.push(shape);
     }
     }
     }
     
     // explode dimension into lines, solids (triangles) and a text entity:
     else if (isDimensionEntity(entity)) {
     shapes = entity.getShapes();
     for (k=0; k<shapes.length; k++) {
     newShapes.push(shapes[k].data());
     }
     var textData = entity.getData().getTextData();
     e = new RTextEntity(entity.getDocument(), textData);
     e.setSelected(true);
     e.copyAttributesFrom(entity.data());
     newEntities.push(e);
     }
     
     // explode leaders into lines and solids (triangles):
     else if (isLeaderEntity(entity)) {
     shapes = entity.getShapes();
     for (k=0; k<shapes.length; k++) {
     newShapes.push(shapes[k].data());
     }
     }
     
     // explode text entities into lines, arcs and splines:
     else if (isTextEntity(entity)) {
     var painterPaths = entity.getPainterPaths();
     for (k=0; k<painterPaths.length; k++) {
     if (painterPaths[k].getFeatureSize()<0) {
     continue;
     }
     shapes = painterPaths[k].getShapes();
     for (n=0; n<shapes.length; n++) {
     shape = shapes[n].data();
     if (isSplineShape(shape)) {
     shape = ShapeAlgorithms.splineToLineOrArc(shape, 0.01);
     }
     if (!isNull(shape)) {
     newShapes.push(shape);
     }
     }
     }
     }
     
     // explode block reference into contained entities:
     else if (isBlockReferenceEntity(entity)) {
     var data = entity.getData();
     var subIds = document.queryBlockEntities(data.getReferencedBlockId());
     for (k=0; k<subIds.length; k++) {
     var subEntity = data.queryEntity(subIds[k]);
     if (isNull(subEntity)) {
     continue;
     }
     
     var e = subEntity.clone();
     storage.setObjectId(e, RObject.INVALID_ID);
     e.setBlockId(document.getCurrentBlockId());
     e.setSelected(true);
     newEntities.push(e);
     }
     }
     
     // add explosion result and delete original:
     if (newShapes.length!==0 || newEntities.length!==0) {
     // add entities based on shapes that resulted from the explosion:
     for (k=0; k<newShapes.length; k++) {
     shape = newShapes[k];
     e = shapeToEntity(entity.getDocument(), shape);
     e.setSelected(true);
     e.copyAttributesFrom(entity.data());
     op.addObject(e, false);
     }
     
     // add entities based on entities that resulted from the explosion:
     for (k=0; k<newEntities.length; k++) {
     op.addObject(newEntities[k], false);
     }
     
     // delete original entity:
     op.deleteObject(entity);
     }
     */
  }
  
  var delta = new RVector(0.0, 0.0);
  var traversed = new Array(newShapes.length);
  for (i=0; i<newShapes.length; i++) traversed[i] = false;
  
  // Now try to find polylines within the entities
  for (i=0; i<newShapes.length; i++) {
    if (traversed[i]==true) continue;
    var sts = newShapes[i];
    traversed[i] = true;
    polyline = new RPolyline();
    var polyStartPoint = sts.getStartPoint();
    var polyEndPoint = sts.getEndPoint();
    polyline.appendVertex(polyStartPoint.operator_add(delta));
    polyline.appendVertex(polyEndPoint.operator_add(delta));
    var again;
    do {
      again = false;
      for (k=0; k<newShapes.length; k++) {
        if (traversed[k]==true) continue;
        var s = newShapes[k];
        if (s.getStartPoint().getDistanceTo(polyEndPoint)<0.01) {
          polyEndPoint = s.getEndPoint();
          polyline.appendVertex(polyEndPoint.operator_add(delta));
          traversed[k] = true; again = true;
        } else if (s.getEndPoint().getDistanceTo(polyEndPoint)<0.01) {
          polyEndPoint = s.getStartPoint();
          polyline.appendVertex(polyEndPoint.operator_add(delta));
          traversed[k] = true; again = true;
        } else if (s.getStartPoint().getDistanceTo(polyStartPoint)<0.01) {
          polyStartPoint = s.getEndPoint();
          polyline.prependVertex(polyStartPoint.operator_add(delta));
          traversed[k] = true; again = true;
        } else if (s.getEndPoint().getDistanceTo(polyStartPoint)<0.01) {
          polyStartPoint = s.getStartPoint();
          polyline.prependVertex(polyStartPoint.operator_add(delta));
          traversed[k] = true; again = true;
        }
      }
    } while (again==true);
    while (polyline.countSegments()>1 && polyline.getStartPoint().getDistanceTo(polyline.getEndPoint())<0.01) {
      polyline.removeLastVertex();
      polyline.setClosed(true);
    }
    newPolylines.push(polyline);
  }
  
  // add entities based on shapes that resulted from the explosion:
  // (there should be none left!)
  for (k=0; k<newShapes.length; k++) {
    if (traversed[k]==false) {
      shape = newShapes[k];
      var shapeEntity = shapeToEntity(document, shape);
      //    e = shapeToEntity(entity.getDocument(), shape);
      shapeEntity.setSelected(true);
      //    e.copyAttributesFrom(entity.data());
      //    op.addObject(e, false);
      op.addObject(shapeEntity, false);
    }
  }

  qWarning("TP 1");
  
  var plBBox = new Array();
  var depth = new Array();
  for (k=0; k<newPolylines.length; k++) {
    plBBox[k] = newPolylines[k].getBoundingBox();
    depth[k] = 0;
  }
  for (i=0; i<newPolylines.length; i++) {
    for (k=0; k<newPolylines.length; k++) {
      if (i!=k) {
        if (plBBox[k].contains(plBBox[i])) {
          depth[i] = depth[i] + 1;
        }
      }
    }
  }
  
  qWarning("TP 2");
  
  // walk through all polylines
  for (k=0; k<newPolylines.length; k++) {
    var pe = newPolylines[k];
    // REntity:
    // virtual bool 	isInside (const RBox &box) const
    // virtual RBox 	getBoundingBox () const
    // setColor(RColor(r,g,g);
    // RBox:
    // bool RBox::contains	(	const RBox & 	other	)	 const
    // bool RBox::isOutside	(	const RBox & 	other	)	 const
    // bool RBox::intersects	(	const RBox & 	other	)	 const
    // RPolyline
    // virtual double 	getDirection1 () const
    // virtual double 	getDirection2 () const
    // bool 	isClosed () const
    /*
    if (pe.isClosed()) {
      if (pe.getDirection1()>0.0)
        pe.setColor(new RColor(0, 255, 0));
      else
        pe.setColor(new RColor(0, 0, 255));
    } else {
      pe.setColor(new RColor(255, 0, 0));
    }
     */
    if (pe.isClosed()) {
      qWarning("Closed PLine " + k);
      // generate an offset curve
      var nSeg = pe.countSegments();
      // is the polyline clockwise or counterclockwise
      var aSum = 0.0;
      /*
      var pVec = pe.getStartPoint().operator_subtract(pe.getEndPoint());
      var qVec = pe.getVertexAt(1).operator_subtract(pe.getVertexAt(0));
      qWarning("  ang " + (pVec.getAngleTo(qVec)/3.141592654*180.0));
      */
      var aVec = pe.getVertexAt(nSeg-2);
      var bVec = pe.getVertexAt(nSeg-1);
      for (i=0; i<nSeg; i++) {
        var cVec = pe.getVertexAt(i);
        var a = aVec.getAngleTo(bVec) - bVec.getAngleTo(cVec);
        a = a/3.1415*180.0;
        if (a>180.0) a-=360.0;
        if (a<-180.0) a+=360.0;
        aSum += a; // aSum is >0 (+360) for clockwise, and <0 (-360) for ccw
        aVec = bVec; bVec = cVec;
      }
      qWarning("aSum="+aSum);
      
      // is this a piece or a hole (even or odd)
      var isOdd = ((depth[i]%2)==1);
      var isCcw = (aSum>0.0);
      if (isCcw==isOdd) {
        pe.reverse(); // may add a vertex!
      }
      
      var np = new RPolyline();
      var ix = nSeg-1;
      var aVec = pe.getVertexAt(nSeg-2);
      var bVec = pe.getVertexAt(nSeg-1);
      for (i=0; i<nSeg; i++) {
        var cVec = pe.getVertexAt(i);
        var a = aVec.getAngleTo(bVec);
        var b = bVec.getAngleTo(cVec);
        /*
        if (b<a) b = b + 2*3.1416;
        var c = a+b/2;
        if (c>=2*3.1416) c = c - 2*3.1416;
        */
        var c = b-a;
        while (c>3.1416) c -= 2*3.1416;
        while (c<-3.1416) c += 2*3.1416;
        qWarning("c=" + c/3.1416*180.0);

        var vix = pe.getVertexAt(ix);
        if (c<-0.51*3.1416) { // angeling away from the cutter in a sharp angle (90..180deg)
          var len = Math.sqrt( 2.0 );
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, a+0.25*3.1416))); //+0.5*3.1416+0.5*c)));
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, b+0.75*3.1416))); // +0.5*3.1416-0.5*c)));
        } else if (c<0.0) { // angeling away from the cutter in a flat angel (0..90deg)
          var len = Math.sqrt( Math.tan(0.5*c)*Math.tan(0.5*c)+1.0*1.0 );
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, a+0.5*3.1416+0.5*c)));
        } else if (c>0.99*3.1416) { // bad angle: too sharp
          np.appendVertex(vix);
        } else if (c>0.51*3.1416) { // angeling toward the cutter in a sharp angle (90..180deg)
          var len = Math.sqrt( Math.tan(0.5*c)*Math.tan(0.5*c)+1.0*1.0 );
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, a+0.5*3.1416+0.5*c)));
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*Math.sqrt(2.0), a+0.5*3.1416+0.5*c)));
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, a+0.5*3.1416+0.5*c)));
        } else { // angeling toward the cutter in a flat angle (0..90deg)
          var len = Math.sqrt( Math.tan(0.5*c)*Math.tan(0.5*c)+1.0*1.0 );
          np.appendVertex(vix.operator_add(RVector.createPolar(1.0*len, a+0.5*3.1416+0.5*c)));
        }

        
        // find the mean angle between these and create an offset vector
        aVec = bVec; bVec = cVec; ix = i;
      }
      np.setClosed(true);
      pe = np;
      
    } else {
      // don't offset the curve
    }
    
    var pEnt = new RPolylineEntity(document, new RPolylineData(pe));
    pEnt.setColor(new RColor(0, 255/(depth[k]+1), 0));
    //pEnt.setSelected(true);
    op.addObject(pEnt, false);
  }
  
  di.applyOperation(op);
  
  this.terminate();
};

